/* $Id:$
***************************************************************************

   X server for LibGGI - Handling of input devices

   Copyright (C) 1997 Jason McMullan	[jmcc@cs.cmu.edu]
   Copyright (C) 1997 Michael Krause	[rawstyle@ms.demo.org]
   Copyright (C) 1998 Marcus Sundberg	[marcus@ggi-project.org]

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

***************************************************************************
*/

#include <stdlib.h>
#include "X11/X.h"
#define NEED_EVENTS
#include "X11/Xproto.h"

#include "scrnintstr.h"
#include "inputstr.h"
#include "mipointer.h"
#include "input.h"
#include "mi.h"

#include "xggi.h"

static void
MouseQueueMovement(DeviceIntPtr device, int dx, int dy, Time tm)
{
	PtrCtrl *pCtrl = &device->ptrfeed->ctrl;

	if ((abs(dx) + abs(dy)) >= pCtrl->threshold) {
		dx = (dx * pCtrl->num) / pCtrl->den;
		dy = (dy * pCtrl->num)/ pCtrl->den;
	}

	miPointerDeltaCursor(dx, dy, tm);
}

static void
xggiQueueMouseEvent(ggi_event *ev)
{
	DevicePtr pMouse;
	xEvent xev;
	static int oldx = 0, oldy = 0;
    
	pMouse = LookupPointerDevice();
	if (!pMouse->on) return;
    
	switch (ev->any.type) {
	case evPtrAbsolute:
		miPointerDeltaCursor(ev->pmove.x - oldx, ev->pmove.y - oldy,
				     TVAL2TIME(ev->any.time));
		oldx = ev->pmove.x;
		oldy = ev->pmove.y;
		return;

	case evPtrRelative:
		MouseQueueMovement((DeviceIntPtr)pMouse,
				   ev->pmove.x, ev->pmove.y,
				   TVAL2TIME(ev->any.time));
		oldx += ev->pmove.x;
		oldy += ev->pmove.y;
		return;

	case evPtrButtonPress:
		xev.u.u.type = ButtonPress;
		break;

	case evPtrButtonRelease:
		xev.u.u.type = ButtonRelease;
		break;
	default:
		/* Unknown pointer event */
		return;
	}

	switch (ev->pbutton.button) {
	case 2:
		/* Right button */
		xev.u.u.detail = 3;
		break;
	case 3:
		/* Middle button */
		xev.u.u.detail = 2;
		break;
	default:
		/* Other button */
		xev.u.u.detail = ev->pbutton.button;
		break;
	}
	xev.u.keyButtonPointer.time = TVAL2TIME(ev->any.time);
	mieqEnqueue(&xev);

	return;
}
  
static int
xggiMouseProc(DeviceIntPtr pDevice, int onoff)
{
	BYTE *map;
	int i;
	DevicePtr pDev = (DevicePtr)pDevice;

	switch (onoff) {
	case DEVICE_INIT:
		if ((map = malloc(sizeof(BYTE)*(xggiScreen.ptrbuttons+1)))
		    == NULL) {
			FatalError("xggiMouseProc: Cannot allocate temporary memory!\n");
		}
		for (i = 0; i <= xggiScreen.ptrbuttons; i++) {
			map[i] = i;
		}
		InitPointerDeviceStruct(pDev, map, xggiScreen.ptrbuttons,
					miPointerGetMotionEvents,
					(PtrCtrlProcPtr)NoopDDA,
					miPointerGetMotionBufferSize());
		free(map);
		break;
	
	case DEVICE_ON:
		pDev->on = TRUE;
		break;
	
	case DEVICE_OFF:
		pDev->on = FALSE;
		break;
	
	case DEVICE_CLOSE:
		break;
	}

	return Success;
}

void
ProcessInputEvents()
{
	mieqProcessInputEvents();
	miPointerUpdate();
}

void
AbortDDX()
{
	ggiExit();
}

static void
xggiDispatchEvents()
{
	ggi_event ev;
	struct timeval ztime = {0,0};
    
	while (ggiEventPoll(MASTER_VIS, emAll, &ztime)) {
		ggiEventRead(MASTER_VIS, &ev, emAll);
		
		switch (ev.any.type) {
		case evPtrRelative:
		case evPtrAbsolute:
		case evPtrButtonPress:
		case evPtrButtonRelease:
			xggiQueueMouseEvent(&ev);
			break;
		case evKeyPress:
		case evKeyRepeat:
		case evKeyRelease:
			xggiQueueKeyboardEvent(&ev);
			break;
		default:
			break;
		}
		ztime.tv_sec = 0;
		ztime.tv_usec = 0;
	}
}

static CARD32
xggiSnarf(OsTimerPtr timer, CARD32 time, pointer data)
{
	xggiDispatchEvents();
	return 20 /* Millis */;
}

static void
xggiInputTimer()
{
	static OsTimerPtr timer;
	timer = TimerSet(NULL, 0, 20 /* Millis */, xggiSnarf, NULL);
}

/*****************************
 * Initailize input handlers *
 *****************************/
void
InitInput(int argc, char *argv[])
{
	DeviceIntPtr p, k;

	ggiSetEventMask(MASTER_VIS, emKeyboard | emPointer);

	p = AddInputDevice(xggiMouseProc, TRUE);
	k = AddInputDevice(xggiKeybdProc, TRUE);
	RegisterPointerDevice(p);
	RegisterKeyboardDevice(k);
	miRegisterPointerDevice(screenInfo.screens[0], p);
	mieqInit((DevicePtr) k, (DevicePtr) p);
	
	xggiInputTimer();
}
